package gui;

import javax.swing.*;
import java.awt.*;
import math.*;
import signalgame.ReputationFunction;


/**
 *  @author Project A.I. Development Team
 */

public class GUI_Function extends JPanel
{

	Displayer       displayer;
	Dimension       size;



    public GUI_Function()
    {
		reset();
		setFunction(null);
    }

	public void reset()
	{
		size = new Dimension(0,0);

	}

	public void setFunction(ReputationFunction function)
	{
			    if (function == null)                                   displayer = new NoFunctionDisplayer();
		else    if (function.fs[0] instanceof DiscreteFunction)         displayer = new DiscreteFunctionDisplayer(function);
		else    if (function.fs[0] instanceof ProbabilisticFunction)    displayer = new ProbabilisticFunctionDisplayer(function);
		else    throw new Error("Unknown Function Type: " + function.getClass());

		Dimension size2 =  displayer.getPreferredSize();
		size = new Dimension(Math.max(size.width, size2.width), Math.max(size.height, size2.height));

		this.setBackground(displayer.getBackgroundColor());

		this.setPreferredSize(size);

		revalidate();
		repaint();
	}


	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		displayer.draw(g);
	}


	abstract class Displayer
	{
		public abstract     Color       getBackgroundColor();
		public abstract     Dimension   getPreferredSize();
		public abstract     void        draw(Graphics g);
	}

	class NoFunctionDisplayer extends Displayer
	{
		public Color        getBackgroundColor()    { return Color.lightGray; }
		public Dimension    getPreferredSize()      { return new Dimension(10,10); }
		public void         draw(Graphics g)        {}
	}

	class DiscreteFunctionDisplayer extends Displayer
	{
		ReputationFunction function;
		public static final int HEIGHT = 20;

		int fromWidth;
		int toWidth;
		int repWidth;
		int elems = 0;

		public DiscreteFunctionDisplayer(ReputationFunction function)
		{
			this.function = function;
			fromWidth = 0;
			toWidth = 0;

			Graphics g = getGraphics();
			FontMetrics fm = g.getFontMetrics();

			for (int j = 0; j < function.fs.length; j++)
			{
				repWidth = Math.max(repWidth, fm.getStringBounds(function.domain.elements[j].toString(), g).getBounds().width);
				DiscreteFunction df = (DiscreteFunction)function.fs[j];

				elems = Math.max(elems, df.domain.elements.length);
				elems = Math.max(elems, df.range.elements.length);

				for (int i = 0; i < df.domain.elements.length; i++)
				{
					fromWidth = Math.max(fromWidth, (int)(fm.getStringBounds(df.domain.elements[i].toString(), g).getWidth()));
				}

				for (int i = 0; i < df.range.elements.length; i++)
				{
					toWidth = Math.max(toWidth, (int)(fm.getStringBounds(df.range.elements[i].toString(), g).getWidth()));
				}
			}

			repWidth += 15;
		}

		public Dimension getPreferredSize()
		{
			Insets border = getInsets();

			return new Dimension(border.left + border.right + repWidth + fromWidth + toWidth + 100,
								 border.top + border.bottom + HEIGHT * elems);
		}

		public Color getBackgroundColor() { return Color.white; }

		public void draw(Graphics g)
		{
			Insets border = getInsets();

			for (int j = 0; j < function.fs.length; j++)
			{
				DiscreteFunction df = (DiscreteFunction)function.fs[j];

				g.drawRect(border.left, border.top + j * elems * HEIGHT, repWidth-5, elems*HEIGHT);

				String name = function.domain.elements[j].toString();
				Rectangle x = g.getFontMetrics().getStringBounds(name, g).getBounds();

				g.drawString(name,  border.left - x.x + 5                   + (repWidth - 10 - x.width)/2,
									border.top  - x.y + j * elems * HEIGHT  + (elems*HEIGHT - x.height)/2);


				for (int i = 0; i < df.domain.elements.length; i++)
				{
					g.drawString(df.domain.elements[i].toString(), border.left + repWidth, border.top + (i + j*elems) *HEIGHT+HEIGHT);
				}

				for (int i = 0; i < df.range.elements.length; i++)
				{
					g.drawString(df.range.elements[i].toString(), border.left + repWidth + fromWidth + 100, border.top + (i + j*elems) *HEIGHT + HEIGHT);
				}

				for (int i = 0; i < df.domain.elements.length; i++)
				{
					int to = df.mapping[i];

					g.drawLine( border.left + + repWidth + fromWidth + 10, border.top + (i + j*elems) *HEIGHT+HEIGHT/2,
								border.left + + repWidth + fromWidth + 90, border.top + (to + j*elems)*HEIGHT+HEIGHT/2);
				}
			}
		}

	}

	class ProbabilisticFunctionDisplayer extends Displayer
	{
		public static final int HEIGHT = 20;

		ReputationFunction function;
		int fromWidth;
		int repWidth;
		int elems = 0;
		int elemHeight;



		public ProbabilisticFunctionDisplayer(ReputationFunction function)
		{
			this.function = function;

			fromWidth = 0;
			Graphics g = getGraphics();
			FontMetrics fm = g.getFontMetrics();


			for (int j = 0; j < function.fs.length; j++)
			{
				repWidth = Math.max(repWidth, fm.getStringBounds(function.domain.elements[j].toString(), g).getBounds().width);
				ProbabilisticFunction pf = (ProbabilisticFunction)function.fs[j];

				elems = Math.max(elems, pf.domain.elements.length);

				for (int i = 0; i < function.domain.elements.length; i++)
				{
					fromWidth = Math.max(fromWidth, (int)(fm.getStringBounds(pf.domain.elements[i].toString(), g).getWidth()));
				}
			}

			repWidth += 15;
			elemHeight = HEIGHT*elems + 5;
		}


		public Color        getBackgroundColor()    { return Color.white; }

		public Dimension getPreferredSize()
		{
			Insets border = getInsets();

			return new Dimension(border.left + border.right + repWidth + fromWidth + 100,
								 border.top + border.bottom + function.domain.elements.length * elemHeight);
		}


		public void draw(Graphics g)
		{
			Insets border = getInsets();
			Dimension s = getSize();

			for (int j = 0; j < function.fs.length; j++)
			{
				ProbabilisticFunction pf = (ProbabilisticFunction)function.fs[j];

				g.drawRect(border.left, border.top + j * elemHeight, repWidth-5, elems*HEIGHT);

				String name = function.domain.elements[j].toString();
				Rectangle x = g.getFontMetrics().getStringBounds(name, g).getBounds();

				g.drawString(name,  border.left - x.x + 5                   + (repWidth   - 10 - x.width )/2,
									border.top  - x.y + j * elemHeight      + (elemHeight -  5 - x.height)/2);


				for (int i = 0; i < pf.domain.elements.length; i++)
				{
					g.drawString(pf.domain.elements[i].toString(), border.left + repWidth, border.top + j * elemHeight + i*HEIGHT+HEIGHT);

					int startx =    repWidth + border.left + fromWidth + 10;
					int starty =    border.top + j * elemHeight + i*HEIGHT + 3;
					int lenx =      s.width - border.right - border.left - fromWidth - repWidth - 20;
					int leny =      HEIGHT-3;

		//			g.drawRect(startx, starty, lenx, leny);

					double cumul = 0;

					for (int k = 0; k < pf.ps[i].domain.elements.length; k++)
					{
						double p = pf.ps[i].ps[k];
						int cumulx = (int)(cumul * lenx);
						int cumull = (int)(lenx * p);
						g.drawRect(startx+cumulx, starty, cumull, leny);

						name = pf.ps[i].domain.elements[k].toString();

						x = g.getFontMetrics().getStringBounds(name, g).getBounds();
						if (x.width < cumull - 5) // text fits
						{
							g.drawString(name, startx + cumulx - x.x + (cumull-x.width)/2, starty - x.y);
						}
						cumul += p;
					}
				}
			}
		}
	}
}