package math;


/** Models a function from a discrete set to a discrete set.
 *  @author Project A.I. Development Team
 */

public class DiscreteFunction implements Function
{
	public final FiniteSet domain;
	public final FiniteSet range;

	public final int [] mapping;

	/** Constructs a total function from 'domain' to 'range'
	 * @param domain The domain
	 * @param range The range
	 * @param mapping The (index based) mapping
	 */

    public DiscreteFunction(FiniteSet domain, FiniteSet range, int [] mapping)
    {
		// Validate mapping
		assert mapping.length == domain.elements.length : "Should be a total function";
		for (int i = 0; i < mapping.length; i++)
		{
			assert mapping[i] < range.elements.length : "Range specification incorrect";
		}

		this.domain = domain;
		this.range = range;
		this.mapping = mapping;
    }


	public double distance(Function F)
	{
		DiscreteFunction df = (DiscreteFunction)F;

		double d = 0;

		for (int i = 0; i < mapping.length; i++)
		{
			if (mapping[i] != df.mapping[i]) d++;
		}
		return d;
	}


	/** Generates a string representation of this funciton on multiple lines
	 *
	 * @return The string
	 */

	public String multiline()
	{
		StringBuffer s = new StringBuffer();
		boolean first = true;
		s.append("{\n");

		for (int i = 0; i < domain.elements.length; i++)
		{
			if (!first)  s.append(",\n");
			else         first = false;

			s.append(domain.elements[i]);
			s.append(" -> ");
			s.append(range.elements[mapping[i]]);
		}

		s.append("}");

		return s.toString();

	}

	/** Generates a string representation of this funciton on one line
	 *
	 * @return The string
	 */

	public String toString()
	{
		StringBuffer s = new StringBuffer();
		boolean first = true;
		s.append("{");

		for (int i = 0; i < domain.elements.length; i++)
		{
			if (!first)  s.append(", ");
			else         first = false;

			s.append(domain.elements[i]);
			s.append(" -> ");
			s.append(range.elements[mapping[i]]);
		}

		s.append("}");

		return s.toString();
	}

	/** Creates a mapping matrix containing the mappings for every possible function
	 *
	 * @param sf the number of elements to map from
	 * @param st the number of elements to map to
	 * @return   the mappint matrix
	 */

	public static int [][] fullFunction(int sf, int st)
	{
		int cardinality =   (int)Math.pow(st, sf);

		int [][] f = new int[cardinality][sf];

		for (int fi = 0; fi < sf; fi++)
		{
			int spans = (int)Math.pow(st, fi);  // the number of spans
			int spanl = cardinality/spans;      // the length of one span
			int spanp = cardinality/st/spans;   // the part of one span that should be set

			for (int ti = 0; ti < st; ti++)
			{
				for (int i1 = 0; i1 < spans; i1++)
				{
					for (int i2 = 0; i2< spanp; i2++)
					{
						f[i1*spanl + ti*spanp + i2][fi] = ti;
					}
				}
			}
		}
		return f;
	}


	/** Computes the set of all total functions from 'from' to 'to'.
	 *  The resulting set contains |from|^|to| elements.
	 *
	 * @param from  The domain of all the computed functions
	 * @param to    The range of all the computed functions
	 * @return      The set of all total functions from 'from' to 'to'
	 */

	public static FiniteSet allPossible(FiniteSet from, FiniteSet to)
	{
		// for every element in from
		// select a mapping to an element in to
		// so |from|^|to| elements in the resulting set

		int [][] fs = fullFunction(from.elements.length, to.elements.length);

		// Create an empty function array
		DiscreteFunction [] f = new DiscreteFunction[fs.length];

		// Fill it with the generated functions
		for (int i = 0; i < fs.length; i++)
		{
		    f[i] = new DiscreteFunction(from, to , fs[i]);
		}

		FiniteSet F = new FiniteSet(f);
		return F;
	}


}