package math;

import java.util.*;


/** Models a mathematical probability distribution over a finite set.
 *  @author Project A.I. Development Team
 */

public class ProbabilityDistribution implements Function
{
	public final FiniteSet  domain;
	public final double []  ps;

	/** Constructs a new Probability Distribution
	 * A probability Distribution is just a function with the followin two restrictions on its values:
	 * 1) The sum of all values must be 1.0
	 * 2) Each value must be in the interval [0.0, 1.0]
	 *
	 * @param domain The domain over which this probabilitydistribution is defined
	 * @param ps     The probability of each domain element
	 *
	 */

    public ProbabilityDistribution(FiniteSet domain, double [] ps)
    {
		assert domain.elements.length == ps.length : "Domain size is " + domain.elements.length + " but there are " + ps.length + " probabilities";
		// Ensure valid
		double p = 0;
		for (int i = 0; i < ps.length; i++)
		{
			assert ps[i] >= 0.0 && ps[i] <= 1.0 : "P(.) is " + ps[i];
		    p += ps[i];
		}

		// Doubles are never truly exact. (So use a small marge)
		double marge = 1E-5;
		assert p > 1.0-marge && p < 1.0+marge : "Sum of probabilities was: " + p;

		this.domain =   domain;
		this.ps =       ps;

    }

	public double distance(Function F)
	{
		ProbabilityDistribution pd = (ProbabilityDistribution)F;

		double d = 0;

		for (int i = 0; i < ps.length; i++)
		{
			d += Math.abs(ps[i] - pd.ps[i]);
		}

		return d/2;
	}


	public Object [] sample (int count)
	{
		Object [] S = new Object[count];

		int         iP =        0;
		double      Pcumul =    ps[iP];

		// Nu rekenen we (logaritmisch (=snel)) een gesorteerde lijst uit van
		// onafhankelijke stochasten getrokken uit de uniforme verdeling over [0,1)

		double logP = 0;

		for (int i = count; i > 0; i--)
		{
			logP += Math.log(Math.random())/i;
			double P = 1-Math.exp(logP);

			// Skip objecten tot we het object hebben dat
			// deze kans heeft
			while (P > Pcumul) Pcumul += ps[++iP];
			S[count-i] = domain.elements[iP];
		}

		return S;
	}


	/** Generates a string representation of this funciton on multiple lines
	 *
	 * @return The string
	 */

	public String multiline()
	{
		StringBuffer s = new StringBuffer();
		boolean first = true;
		s.append("{\n");

		for (int i = 0; i < domain.elements.length; i++)
		{
			if (!first)  s.append(",\n");
			else         first = false;

			s.append(domain.elements[i]);
			s.append(" -> ");
			s.append(ps[i]);
		}

		s.append("}");

		return s.toString();

	}

	/** Generates a string representation of this funciton on one line
	 *
	 * @return The string
	 */

	public String toString()
	{
		StringBuffer s = new StringBuffer();
		boolean first = true;
		s.append("{");

		for (int i = 0; i < domain.elements.length; i++)
		{
			if (!first)  s.append(", ");
			else         first = false;

			s.append(domain.elements[i]);
			s.append(" -> ");
			s.append(ps[i]);
		}

		s.append("}");

		return s.toString();
	}



	/** Creates a truly uniform distribution over s
	 *
	 * @param s The set over which the distribution must be computed
	 * @return  The new universal distribution over this set
	 */

	public static ProbabilityDistribution uniform(FiniteSet s)
	{
		assert s.elements.length > 0 : "Uniform distribution requires at least one element";

		double [] ps = new double[s.elements.length];

		double p = 1.0/s.elements.length;

		for (int i = 0; i < ps.length; i++)
		{
			ps[i] = p;
		}

		return new ProbabilityDistribution(s, ps);
	}

	/** Creates a random uniform-like distribution over s
	 *
	 * @param s The set over which this distribution must be created
	 * @return The new probability distribution
	 */


	public static ProbabilityDistribution random(FiniteSet s)
	{
		assert s.elements.length > 0 : "Random distribution requires at least one element";

		double [] ps = new double[s.elements.length];

		// First give everybody one point
		for (int i = 0; i < s.elements.length; i++)
		{
			ps[i]++;
		}

		// Then divide 9 times |set| points at random
		for (int i = 0; i < s.elements.length*9; i++)
		{
			ps[(int)(Math.random()*s.elements.length)]++;
		}

		// 10 * |set| points were distributed, normalize
		for (int i = 0; i < ps.length; i++)
		{
			ps[i] /= 10*s.elements.length;
		}

		return new ProbabilityDistribution(s, ps);
	}

	public static ProbabilityDistribution justOne(FiniteSet s)
	{
		double [] ps = new double[s.elements.length];
		ps[(int)(Math.random() * ps.length)] = 1.0;
		return new ProbabilityDistribution(s, ps);
	}





}