package signalgame;

import math.*;
import java.util.*;


/** The Dynamic Information during the Singal Game Simulation
 */

public abstract class DynamicInformation
{
	public final IIndividual []          I;
	public final ProbabilityDistribution Pi;
	public final double []               Ei;
	public final double                  E;
	public final StaticInformation       S;

	public DynamicInformation(StaticInformation S, IIndividual [] I, ProbabilityDistribution Pi, boolean sort)
	{
		this.S =                S;
		final double [] Ei =    calculateExpectedUtility(I,  Pi);
		E =                     calculateMeanUtility    (Ei, Pi);

		if (sort)
		{
			Integer [] is = new Integer[I.length];
			for (int i = 0; i < is.length; i++) is[i] = new Integer(i);

			// Sort the population indices to increasing utility
			Arrays.sort(is, new Comparator()
				{
					public int compare(Object o1, Object o2)
					{
						assert(o1 instanceof Integer && o2 instanceof Integer);
						int i1 = ((Integer)o1).intValue();
						int i2 = ((Integer)o2).intValue();

						// Compare reversed, to sort hight to low
						return Double.compare(Ei[i2], Ei[i1]);
					}
				}
			);

			this.I =        new IIndividual[I.length];
			double [] ps =  new double [Pi.domain.elements.length];
			this.Ei =       new double [Ei.length];

			for (int i = 0; i < is.length; i++)
			{
				int id =        ((Integer)is[i]).intValue();
				this.I[i] =     I[id];
				ps[i] =         Pi.ps[id];
				this.Ei[i] =    Ei[id];
			}

			this.Pi = new ProbabilityDistribution(new FiniteSet(this.I, true), ps);
		}
		else
		{
			this.I =    I;
	    	this.Pi =   Pi;
			this.Ei =   Ei;
		}
	}


	public abstract double [] calculateExpectedUtility(IIndividual [] I, ProbabilityDistribution Pi);

	/** Calculates the expected expected utility
	 *
	 * @return Utility
	 */

	public double calculateMeanUtility(double [] Ei, ProbabilityDistribution Pi)
	{
		double E = 0;

		// Now calculate expected total utility
		for (int i = 0; i < Ei.length; i++)
		{
			E += Pi.ps[i] * Ei[i];
		}

		return E;
	}


	public abstract DynamicInformation   evolve();
	public abstract String               multiline();


	public IIndividual getMostUtilous()
	{
		int bestSi = 0;
		for (int i = 1; i < Ei.length; i++)
		{
			if(Ei[i] > Ei[bestSi]) bestSi = i;
		}
		return I[bestSi];
	}

}