package signalgame;

import java.util.HashMap;

import math.*;
import util.*;

/** A 3 dimensional function of type: set1 X set2 X {speaker, hearer} -> double.
 *  @author Project A.I. Development Team
 */

public class PayoffMatrix
{
	public final FiniteSet set1;
	public final FiniteSet set2;
	public static final FiniteSet set3 = new FiniteSet(new Object [] {"Speaker", "Hearer"});
	public final double [][][] values;

	/** Constructs a new payoff (utility) matrix. The parameters are validated.
	 *  The cardinality of set1 should be equal to the size of the first dimension in 'values'.
	 *  The cardinality of set2 should be equal to the size of the second dimension in 'values'.
	 *  The size of the third dimension in 'values' should be 2. (One value for the speaker, one for the hearer)
	 *
	 * @param set1      Set1.
	 * @param set2      Set2.
	 * @param values    The utilities.
	 */

    public PayoffMatrix(FiniteSet set1, FiniteSet set2, double [][][] values)
    {
		// Check the values structure for consistency with row and col labels
		assert values.length == set1.elements.length : "1st dimension of Values (" + values.length + ") should be as large as |set1| (" + set1.elements.length +")";

		for (int si = 0; si < set1.elements.length; si++)
		{
			assert values[si].length == set2.elements.length : "2nd dimension of Values (" + values[si].length + ") should be as large as |set2| (" + set2.elements.length +")";

			for (int ai = 0; ai < set2.elements.length; ai++)
			{
				assert values[si][ai].length == set3.elements.length : "3rd dimension of Values (" + values[si][ai].length + ")should be as large as |set3| (" + set3.elements.length +")";
			}
		}

		this.set1 =     set1;
		this.set2 =     set2;
		this.values =   values;
    }

	/** Generates a nice string representation of this funciton on multiple lines.
	 *  The string layout is computed in such a way that it always fits.
	 *  The column headers are centered above the columns
	 *
	 * @return The string
	 */


	public String multiline()
	{
		StringBuffer s = new StringBuffer();

		int labelLength = 0;
		int columnLength = 0;

		for (int r = 0; r < set1.elements.length; r++)
		{
			labelLength = Math.max(labelLength, set1.elements[r].toString().length());
		}

		for (int c = 0; c < set2.elements.length; c++)
		{
			columnLength = Math.max(columnLength, set2.elements[c].toString().length());
		}

		for (int r = 0; r < set1.elements.length; r++)
		{
			for (int c = 0; c < set2.elements.length; c++)
			{
				for (int v = 0; v < set3.elements.length; v++)
				{
					columnLength = Math.max(columnLength, Double.toString(values[r][c][v]).length());
				}
			}
		}


		s.append("\n");
		for (int pad = 0; pad < labelLength + 3; pad++)
		{
			s.append(' ');
		}

		for (int c = 0; c < set2.elements.length; c++)
		{
			String l = set2.elements[c].toString();

			int pre = ((columnLength*2+6)-l.length())/2;

			for (int pad = 0; pad < pre; pad++)
			{
				s.append(' ');
			}
			s.append(l);
			for (int pad = pre + l.length(); pad < columnLength*2+6; pad++)
			{
				s.append(' ');
			}
		}

		for (int r = 0; r < set1.elements.length; r++)
		{
			s.append("\n");

			String l = set1.elements[r].toString();
			s.append(l);
			for (int pad = l.length(); pad < labelLength+2; pad++)
			{
				s.append(' ');
			}

			s.append(" | ");
			for (int c = 0; c < set2.elements.length; c++)
			{
				if (c > 0) s.append(" | ");
				l = Double.toString(values[r][c][0]);
				s.append(l);
				for (int pad = l.length(); pad < columnLength; pad++)
	    		{
		    		s.append(' ');
				}
				s.append(" / ");
				l = Double.toString(values[r][c][1]);
				s.append(l);
				for (int pad = l.length(); pad < columnLength; pad++)
	    		{
		    		s.append(' ');
				}
			}
			s.append(" |");
		}
		return s.toString();
	}

	/** Generates a string representation of this funciton
	 *  This function just calls 'multiline', because a one-dimensional representation
	 *  of a Matrix is neither readable nor useful.
	 *
	 * @return The string
	 */


	public String toString()
	{
		return multiline();
	}
}
