package signalgame.instancebased;

import math.*;
import util.*;
import signalgame.*;


/** The dynamic information in a population based Signal Game.
 *  @author Project A.I. Development Team
 */


public class InstanceBasedDynamic extends DynamicInformation
{
	public static final boolean DEBUG = false;


    public InstanceBasedDynamic(StaticInformation p, IIndividual [] I, ProbabilityDistribution Pi)
    {
		super(p, I, Pi, true);
		assert p instanceof InstanceBasedStatic;
		assert I instanceof Individual [];
    }

	public String toString()
	{
		return this.multiline();
	}


	public String multiline()
	{
		StringBuffer s = new StringBuffer();

		for (int i = 0; i < I.length; i++)
		{
			if (i != 0) s.append('\n');
			s.append(Formatter.padr(i, 20));
			s.append(Formatter.padr(Ei[i], 20));
		}
		return s.toString();
	}



	public double [][] distanceTable(IIndividual [] I, ProbabilityDistribution Pi)
	{
		double [][] PT = new double[I.length][I.length];
		double [] sum = new double[I.length];

		InstanceBasedStatic ibs = (InstanceBasedStatic)S;

		for (int i = 0; i < I.length; i++)
		{
			Individual Ii = (Individual)I[i];
			for (int j = i; j < I.length; j++)
			{
				Individual Ij = (Individual)I[j];

				double d = Ii.distance(Ij); // symmetric!!
				double p = Math.exp(-ibs.correlation * d*d);
				PT[i][j] = p;
				sum[i] += p;

				if (i != j) //prevent double counts
				{
					PT[j][i] = p;
					sum[j] += p;
				}
			}
		}

		if (DEBUG)
		{
			for (int i = 0; i < I.length; i++)
			{
				for (int j = 0; j < I.length; j++)
				{
					System.out.print((int)(100*PT[i][j])/100.0 + "\t");
				}
				System.out.println("\tsum: " + sum[i]);
			}
			System.out.println();
		}

		for (int i = 0; i < I.length; i++)
		{
			for (int j = 0; j < I.length; j++)
			{
				PT[i][j] /= sum[i];
			}
		}

		if (DEBUG)
		{
			for (int i = 0; i < I.length; i++)
			{
				for (int j = 0; j < I.length; j++)
				{
					System.out.print((int)(100*PT[i][j])/100.0 + "\t");
				}
				System.out.println();
			}
			System.out.println();
		}

		return PT;
	}




	public double [] calculateExpectedUtility(IIndividual [] I, ProbabilityDistribution Pi)
	{
		InstanceBasedStatic p = (InstanceBasedStatic)S;

		double [][] PT = distanceTable(I, Pi);

		double [] Ei = new double[I.length];

		// iterate over all individuals as speakers
		for (int i = 0; i < I.length; i++)
		{
			// Calculate the probability that i speaks
			// P(speaker=i)
//			double P_si = Pi.ps[i]/2;
			Individual Ii = (Individual)I[i];
//			if (DEBUG) System.out.println("P(speaker=" + i + ") = " + P_si);

			for (int j = 0; j < I.length; j++)
			{
				// Update the probability with the fact that j is the hearer
				double P_hj_given_si = PT[i][j]/2; //Pi.ps[j]/2;
				double P_si_given_hj = PT[j][i]/2; // Bayes Rule Contracted
//				if (DEBUG) System.out.println("P(speaker=" + i + ", hearer=" + j + ") = " + P_si_hj);


				Individual Ij = (Individual)I[j];

				ProbabilisticFunction Ss = (ProbabilisticFunction)Ii.Ss.fs[I[j].reputationID];
				ProbabilisticFunction Sh = (ProbabilisticFunction)Ij.Sh.fs[I[i].reputationID];

				// iterate over all states the world can be in
				for (int t = 0; t < p.T.elements.length; t++)
				{
					// Update the probability with the fact that the world state is t
					// P(speaker=i, hearer=j, state=t) = P(state=t | speaker=i, hearer=j) * P(speaker=i, hearer=j)
					double P_t = p.Pt.ps[t];
//					if (DEBUG) System.out.println("P(speaker=" + i + ", hearer=" + j + ", state=" + t + ") = " + P_si_hj_t);

					// iterate over all messages the speaker (i) can say
					for (int m = 0; m < p.M.elements.length; m++)
					{
						// Update the probability with the fact that the message m is sent
						// P(speaker=i, hearer=j, state=t, message=m) = P(message=m | speaker=i, hearer=j, state=t) * P(speaker=i, hearer=j, state=t)
						double P_t_m = Ss.ps[t].ps[m] * P_t;
//						if (DEBUG) System.out.println(Formatter.padr("P(speaker=" + i + ", hearer=" + j + ", state=" + t + ", message=" + m + ") = " + P_si_hj_t_m, 80) +
//													  Formatter.padr("Message utility = ", 30) + Formatter.padl(p.Costs.values[t][m][0],8) + "/ " + Formatter.padr(p.Costs.values[t][m][1], 8));

						double [] Mcomm = p.Costs.matrices[I[i].reputationID][I[j].reputationID].values[t][m];
						Ei[i] += P_hj_given_si * P_t_m * Mcomm[0];
						Ei[j] += P_si_given_hj * P_t_m * Mcomm[1];

						// iterate over all actions the hearer (j) can perform
						for (int a = 0; a < p.A.elements.length; a++)
						{
							// Update the probability with the fact that action a is performed
							// P(speaker=i, hearer=j, state=t, message=m, action=a) = P(action=a | speaker=i, hearer=j, state=t, message=m) * P(speaker=i, hearer=j, state=t, message=m)
							double P_t_m_a = Sh.ps[m].ps[a] * P_t_m;
//							if (DEBUG) System.out.println(Formatter.padr("P(speaker=" + i + ", hearer=" + j + ", state=" + t + ", message=" + m + ", action=" + a +") = " + P_si_hj_t_m_a, 80) +
//														  Formatter.padr("State-action utility = ", 30) + Formatter.padl(p.Utilities.values[t][a][0], 8) + "/ " + Formatter.padr(p.Utilities.values[t][a][1], 8));

						    double [] Mutil = p.Utilities.matrices[I[i].reputationID][I[j].reputationID].values[t][a];
						    Ei[i] += P_hj_given_si * P_t_m_a * Mutil[0];
							Ei[j] += P_si_given_hj * P_t_m_a * Mutil[1];
						}
					}
				}
			}
		}
		return Ei;
	}


	public DynamicInformation evolve()
	{
//		InstanceBasedStatic p = (InstanceBasedStatic)S;
//		return p.Es.update(this);

		InstanceBasedStatic p = (InstanceBasedStatic)S;

		ProbabilityDistribution Pnew = S.dynamics.evolve(Ei, E, Pi);

		Individual [] Imut = new Individual[I.length];
		Object [] Inew =     Pnew.sample(Imut.length);

		for (int i = 0; i < Imut.length; i++)
		{
			Imut[i] = (Individual)Inew[i];
			if (Math.random() < p.mutationrate) Imut[i] = p.Es.mutate(Imut[i], S.mutator);
		}

		ProbabilityDistribution Pmut = ProbabilityDistribution.uniform(new FiniteSet(Imut, true));

		return new InstanceBasedDynamic(S, Imut, Pmut);
	}



	public IIndividual getMostUtilous()
	{
		int bestSi = 0;
		for (int i = 1; i < Ei.length; i++)
		{
			if(Ei[i] > Ei[bestSi]) bestSi = i;
		}

		return I[bestSi];
	}

}
